
import io
import pickle
from contextlib import closing
import click
from matplotlib import pyplot as plt
import networkx as nx
import pydot
from sqlalchemy import create_engine, MetaData
from sqlalchemy.orm import Session 
from views_query_planning import query_planning, compose_join, query_with_ops

@click.group(name = "vqp")
def vqp():
    """
    ViEWS query planning
    ====================

    This CLI exposes core functionality from the ViEWS query planner, making it
    possible to inspect and understand the query planning operations taking
    place with the library.

    Examples:

        > Generate a join network and output it to a plot
        vqp with postgresql://user@database/dbname join-network - | vqp dot-to-plot - out.png

    """
    pass

@vqp.group(name = "with")
@click.argument("connection-string")
@click.option("--schema", default = "public")
@click.pass_context
def with_network(ctx: click.Context, connection_string: str, schema: str):
    ctx.ensure_object(dict)
    ctx.obj["engine"] = create_engine(connection_string)
    ctx.obj["session"] = Session(bind = ctx.obj["engine"])

    metadata = MetaData(bind = ctx.obj["engine"], schema = schema)
    metadata.reflect()

    ctx.obj["metadata"] = metadata
    ctx.obj["network"] = query_planning.join_network(ctx.obj["metadata"].tables)
    ctx.obj["schema"] = schema

@with_network.command(name = "join-network")
@click.pass_context
@click.argument("file", type = click.File("w"))
def join_network(ctx: click.Context, file: io.BufferedWriter):
    file.write(nx.nx_pydot.to_pydot(ctx.obj["network"]).to_string())

@with_network.command(name = "query")
@click.pass_context
@click.argument("level-of-analysis", type = str)
@click.argument("table", type = str)
@click.argument("column-name", type = str)
@click.argument("file", type = click.File("w"))
@click.option("-a","--aggregation-function", type = str, default = "avg")
def query(
        ctx: click.Context, 
        level_of_analysis: str,
        table: str,
        column_name: str,
        aggregation_function: str,
        file: io.BufferedWriter):
    
    index_columns = ["id"]

    query = query_with_ops(
            ctx.obj["session"].query(),
            compose_join,
            ctx.obj["network"],
            level_of_analysis,
            table,
            column_name,
            index_columns,
            aggregation_function
            )

    click.echo(str(query))

@with_network.command(name = "dump")
@click.argument("file", type = click.File("wb"))
@click.pass_context
def dump_db_reflection(ctx: click.Context, file: io.BufferedWriter):
    pickle.dump(ctx.obj["metadata"], file)

@vqp.command(name = "dot-to-plot")
@click.argument("dotfile", type = click.File("r"))
@click.argument("outfile", type = click.File("wb"))
def dot_to_plot(dotfile = io.BufferedReader, outfile = io.BufferedWriter):
    data = dotfile.read()
    click.echo(data)
    pydot_graph = pydot.graph_from_dot_data(data)
    if pydot_graph is None:
        click.echo("failed to read dot graph from data:")
        click.echo(data)
        return 1
    else:
        pydot_graph, *_ = pydot_graph

    graph = nx.nx_pydot.from_pydot(pydot_graph)
    fig,ax = plt.subplots(figsize = (20,20))
    nx.draw(graph, with_labels = True, ax = ax)
    fig.savefig(outfile)






"""
querying
========

This module contains code for computing SQL queries. Functionality is primarily
exposed via the QueryComposer class.

"""
import logging
from typing import Deque, Tuple, TypeVar
from collections import deque
from sqlalchemy import Table, Column
from sqlalchemy import sql
from sqlalchemy.sql.selectable import Select, Join
from toolz.functoolz import curry
from networkx import NetworkXNoPath, DiGraph
from networkx.algorithms.shortest_paths import shortest_path
from pymonad.either import Left, Right, Either
from pymonad.maybe import Maybe, Just, Nothing

logger = logging.getLogger(__name__)

T = TypeVar("T")

class QueryComposer():

    def __init__(self,
            network: DiGraph,
            loa_name: str,
            time_index: str,
            unit_index: str,
            outer: bool = False):
        """
        QueryComposer
        =============

        arguments:
            network (networkx.DiGraph): A network as produced by views_query_planning.join_network
            loa_name (str):             The name of the table to use as the level of analysis.
            time_index (str):           The name of the column in loa_name to use as the time-index
            unit_index (str):           The name of the column in loa_name to use as the unit-index
            outer (bool):               Should expressions outer-join to keep all rows in at the LOA?

        The query composer lets you compute SQL queries for retrieving data in
        a normalized database, through joining and aggregating.

        This is exposed through the .expression method, which lets you get an
        expression selecting a column from a table, at the level of analysis
        that the composer was instantiated with.

        """
        self.network = network
        self._tables = {tbl.name: tbl for tbl in self.network}
        self.loa_name = loa_name
        self._time_index = time_index
        self._unit_index = unit_index
        self._isouter = outer

    def joins(self, join_path: Deque[Table])-> Join:
        """
        joins
        =====

        arguments:
            join_path (Deque[Table]): A deque of potentially joinable tables

        returns:
            Join: A join expression that can be used for futher query composition.

        """
        a = join_path.popleft()
        expression = a
        for b in join_path:
            try:
                edge = self.network[a][b]
                condition = (edge["reference"] == edge["referent"])
            except KeyError:
                edge = self.network[b][a]
                condition = (edge["referent"] == edge["reference"])

            expression = sql.join(expression, b, condition, isouter = self._isouter)
            a = b
        return expression


    def expression(self, table: str, column: str, aggregation_function: str = "sum")-> Either[str, str]:
        """
        expression
        ==========

        arguments:
            table (str):                The name of a table to join to
            column (str):               The name of a column to select from the table.
            aggregation_function (str): The name of a function to use to aggregate if applicable.

        returns:
            Either[str, str]

        Computes an SQL expression that joins and selects a desired
        table.column to the self.level_of_analysis, including the LOAs index
        columns.

        The expression will aggregate, grouping on the LOA table and applying
        the aggregation function, if necessary.

        Requires there to be _a_ valid join-path between table and LOA. This
        means that to join table_a and table_b there _either_ needs to be a
        relationship like this:

        (one-to-many)
        ┌───────┐     ┌───────┐     ┌───────┐
        │table_a│─fk─►│table_b│─fk─►│table_c│
        └───────┘     └───────┘     └───────┘

        Or a relationship like this:

        (aggregates, many-to-one)
        ┌───────┐     ┌───────┐     ┌───────┐
        │table_a│◄─fk─│table_b│◄─fk─│table_c│
        └───────┘     └───────┘     └───────┘

        Across any number of intermediate tables.

        """

        to_select = (self._column(table,column)
                .maybe(Left(f"Column {table}.{column} doesn't exist"), Right)
                .then(lambda c: c.label(c.name)))

        index_columns = (self.index_columns
                .maybe(Left("Index columns not found."), Right)
                .then(lambda columns: (c.label(c.name) for c in columns)))

        tables = [self.loa_table, self._table(table)]

        forwards_path = Maybe.apply(curry(path, True)).to_arguments(Just(self.network), *tables).join()

        join_path, aggregates = forwards_path.maybe(
                (Maybe.apply(curry(path, False)).to_arguments(Just(self.network), *tables).join(), True),
                lambda x: (Just(x), False))

        joins = (join_path.maybe(Left("Failed to find path"), Right)
            .then(self.joins))

        select_from = (Either.apply(curry(lambda idx_col, joins: sql.select(*idx_col).select_from(joins)))
            .to_arguments(index_columns, joins))

        if aggregates:
            selection_function = curry(aggregate, aggregation_function, self.loa_table.value)
        else:
            selection_function = lambda to_select, select_from: Right(select_from.add_columns(to_select))

        return Either.apply(curry(selection_function)).to_arguments(to_select, select_from).join().then(str)

    def _table(self, name)-> Maybe[Table]:
        """
        _table
        ======

        arguments:
            name (str): name of the table to get

        returns:
            Maybe[Table]: table, if present.

        """
        try:
            return Just(self._tables[name])
        except KeyError:
            return Nothing

    def _column(self, table, column)-> Maybe[Column]:
        """
        _column
        ======

        arguments:
            table (str):  name of the table to get
            column (str): name of the column to get

        returns:
            Maybe[Column]: column, if table and column are present.

        """
        def get_key(lookup, key):
            try:
                return Just(lookup[key])
            except KeyError:
                return Nothing

        return self._table(table).then(lambda tbl: get_key(tbl.c, column))

    @property
    def loa_table(self) -> Maybe[Table]:
        return self._table(self.loa_name)

    @property
    def index_columns(self) -> Maybe[Tuple[Column, Column]]:
        time, unit = (self._column(self.loa_name, col) for col in (self._time_index, self._unit_index))
        return Maybe.apply(curry(lambda t,u: Just((t,u)))).to_arguments(time, unit).join()

def path(forwards: bool, network: DiGraph, a: T, b: T)-> Maybe[Deque[T]]:
    """
    path
    ====

    arguments:
        forwards (bool):               Find path with or against direction in network.
        network (networkx.Digraph[T]): A network of nodes of type T to find a path in
        a (T):                         A node in the network
        b (T):                         A node in the network

    returns:
        Maybe[Deque[T]]: A deque of steps from a to b

    """

    if forwards:
        x,y = a,b
        post = lambda x:x
    else:
        x,y = b,a
        post = reversed

    try:
        return Just(deque(post(shortest_path(network, x, y))))
    except NetworkXNoPath:
        logger.debug(f"No path between {a} and {b} in {network}! ({'forwards' if forwards else 'backwards'})")
        return Nothing

def aggregate(aggregation_function_name: str, aggregate_to: Table, column: Column, select: Select)-> Select:
    """
    aggregate
    =========

    arguments:
        aggregation_function_name (str): Name of agg. fn
        aggregate_to (Table):            Table to aggregate to
        column (Column):                 Column to aggregate
        select (Select):                 Select statement to mutate

    returns:
        Select: Select statement with added select of aggregated column.

    Adds selection of an aggregated column to a select statement.

    """

    _allowed_aggregation_functions = [
            "sum",
            "max",
            "min",
            "avg",
        ]
    if aggregation_function_name in _allowed_aggregation_functions:
        aggregation_function = getattr(sql.func, aggregation_function_name)
        return Right(select
                .add_columns(aggregation_function(column).label(column.name + "_" + aggregation_function_name))
                .group_by(*set(aggregate_to.primary_key)))
    else:
        return Left(f"Aggregation function {aggregation_function_name} does not exist.")



import logging
from typing import Dict
import enum
from itertools import chain

import networkx as nx
from networkx.exception import NetworkXNoPath
from networkx.algorithms.shortest_paths import shortest_path

import sqlalchemy as sa

from . import exceptions, definitions

logger = logging.getLogger(__name__)

def class_partial(method,*args,**kwargs):
    def undotted(instance):
        return getattr(instance,method)(*args,**kwargs)
    return undotted

class Direction(enum.Enum):
    forwards = 1
    backwards = 2
    equal = 3

def join_network(tables:Dict[str,sa.Table])->nx.DiGraph:
    """
    Creates a directed graph of the FK->Referent relationships present in a
    list of tables. This graph can then be traversed to figure out how to
    perform a join when retrieving data from this collection of tables.
    """
    digraph = nx.DiGraph()
    def get_fk_tables(fk):
        return [tbl for tbl in tables.values() if fk.references(tbl)]

    for table in tables.values():
        for fk in table.foreign_keys:
            try:
                ref_table = get_fk_tables(fk)[-1]
            except IndexError:
                logger.debug("No table found for fk: %s",str(fk))
                continue

            digraph.add_edge(
                    table,
                    ref_table,
                    reference=fk.parent,
                    referent=fk.column
                )

            if {fk.parent} == set(table.primary_key):
                digraph.add_edge(
                        ref_table,
                        table,
                        reference=fk.column,
                        referent=fk.parent,
                    )

    logger.debug("Database is a digraph with %s nodes",len(digraph.nodes))
    return digraph

def compose_join(network,loa_name,table_name,column_name,loa_index_columns,agg_fn="avg"):
    """
    compose_join

    Creates a list of operations that can be applied to a Query object, making
    it retrieve data at a certain LOA

    :param network: A networkx graph representing a database
    :param loa_name: The name of the LOA to retrieve data at
    :param table_name: The name of the table from which to retrieve the data 
    :param column_name: The name of the column to retrieve 
    :param agg_fn: If the retrieval op. needs to aggregate data, if will do so with this function
    :returns: An iterator containing functions to be applied to a Query object 
    :raises QueryError: If table_name.column_name does not exist in the DB 
    :raises ConfigError: If the requested loa is not configured  
    """

    lookup = {tbl.name:tbl for tbl in network}
    get_col_ref = lambda tbl,name: lookup[tbl].c[name]#.label(tbl+"_"+name)

    loa_table = lookup[loa_name]

    try:
        column_ref = get_col_ref(table_name,column_name)
    except KeyError as ke:
        raise exceptions.QueryError(f"{table_name}.{column_name} not found") from ke

    index_columns_ref = []
    for idx_table_name,idx_column_name in loa_index_columns:
        try:
            index_columns_ref.append(get_col_ref(idx_table_name,idx_column_name))
        except KeyError as ke:
            raise exceptions.ConfigError(f"Index column for {loa_table}"
                    f" {idx_table_name}.{idx_column_name} not found")

    all_columns = list(set(index_columns_ref+[column_ref]))
    names = [c.table.name + "_" + c.name for c in all_columns]
    aggregates = False


    # Compute joins (+ agg)
    join = []
    joined = set()
    for idx,c in enumerate(all_columns):
        try:
            path = shortest_path(network, loa_table, c.table)
        except NetworkXNoPath:
            try:
                assert agg_fn in definitions.AGGREGATION_FUNCTIONS
            except AssertionError:
                raise exceptions.AggregationNameError(
                        f"Aggregation function {agg_fn} "
                        "is not available. "
                        "Available functions are: "
                        f"{', '.join(definitions.AGGREGATION_FUNCTIONS)}."
                        )

            aggregates = True
            path = shortest_path(network,c.table, loa_table)
            path.reverse()
            all_columns[idx] = getattr(sa.func,agg_fn)(c)
            names[idx] = names[idx]+"_"+agg_fn

        prev = loa_table
        for table in path[1:]:
            if table not in joined:
                try:
                    con = network[prev][table]
                    join.append(class_partial("join",table,con["reference"]==con["referent"]))
                except KeyError:
                    con = network[table][prev]
                    join.append(class_partial("join",table,con["referent"]==con["reference"]))
                joined = joined.union({table})
            else:
                pass
            prev = table

    # Selects
    all_columns = [col.label(name) for col,name in zip(all_columns,names)]
    select = [class_partial("add_columns",col) for col in all_columns]

    # Aggregation
    if aggregates:
        group_by = [class_partial("group_by",c) for c in index_columns_ref]
    else:
        group_by = []

    logger.debug("%s selects",len(select))
    logger.debug("%s joins",len(join))
    logger.debug("%s groupbys",len(group_by))

    return chain(select,[class_partial("select_from",loa_table)],join,group_by)

def query_with_ops(query,op_composer,*args,**kwargs):
    for op in op_composer(*args,**kwargs):
        query = op(query)
    return query



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['views_query_planning']

package_data = \
{'': ['*']}

install_requires = \
['PyMonad>=2.4.0,<3.0.0',
 'SQLAlchemy>=1.4.27,<2.0.0',
 'click>=8.0.3,<9.0.0',
 'matplotlib>=3.5.0,<4.0.0',
 'networkx>=2.6.3,<3.0.0',
 'psycopg2>=2.9.2,<3.0.0',
 'pydot>=1.4.2,<2.0.0',
 'toolz>=0.11.2,<0.12.0']

entry_points = \
{'console_scripts': ['vqp = views_query_planning.cli:vqp']}

setup_kwargs = {
    'name': 'views-query-planning',
    'version': '1.3.0',
    'description': '',
    'long_description': None,
    'author': 'peder2911',
    'author_email': 'pglandsverk@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
